import numpy as np
from sklearn.utils import indexable
from sklearn.utils.validation import _num_samples


def walk_forward_release(X, y, releases):
    """
    Generate train and test splits fro TimeSeriesSplit on releases.
    Train consists of a release or a list of successive releases, and
    the test set consist of the next release in time

    :param X: array-like of shape (n_samples, m_features)
    :param y: array-like of shape (n_samples,)
    :param releases : array-like of shape (n_samples,)
        Group labels for the samples used while splitting the dataset into
        train/test set.
        Must be an ordered list of integer, i.e., [1, 1, 1, 2, 2, 3, 4, 4, etc.].
        Each integer denote a given release. Files within the same release have the same
        group id.
    """
    X, _, releases = indexable(X, y, releases)
    n_samples = _num_samples(X)
    n_folds = len(set(releases))  # Number of distinct groups (releases)
    if n_folds > n_samples:
        raise ValueError(
            ("Cannot have number of folds ={0} greater"
             " than the number of samples: {1}.").format(n_folds, n_samples))

    indices = np.arange(n_samples)
    offset = 0

    for _ in range(0, n_folds - 1):
        train_indices = [i for i, x in enumerate(releases) if x == releases[offset]]
        offset += len(train_indices)
        test_indices = [i for i, x in enumerate(releases) if x == releases[offset]]

        yield indices[:offset], indices[offset: offset + len(test_indices)]



from imblearn.under_sampling import RandomUnderSampler
from imblearn.over_sampling import RandomOverSampler

balancers_map = dict(
    none=None,
    rus=RandomUnderSampler(sampling_strategy='majority', random_state=42),
    ros=RandomOverSampler(sampling_strategy='minority', random_state=42)
)



from sklearn.ensemble import RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB
from sklearn.svm import SVC

classifiers_map = {
    'dt': DecisionTreeClassifier(),
    'nb': GaussianNB(),
    'logit': LogisticRegression(),
    'rf': RandomForestClassifier(),
    'svm': SVC()
}



import io
import numpy as np
import os
import pandas as pd
import joblib
import json

from typing import List

from imblearn.pipeline import Pipeline
from sklearn.metrics import make_scorer, matthews_corrcoef
from sklearn.feature_selection import VarianceThreshold
from sklearn.model_selection import RandomizedSearchCV

from .balancers import balancers_map
from .classifiers import classifiers_map
from .hyper_parameters import search_params_map
from .normalizers import normalizers_map
from .model_validation import walk_forward_release


def prepare_training_data(data: pd.DataFrame):
    assert 'failure_prone' in data.columns
    assert 'commit' in data.columns
    assert 'committed_at' in data.columns
    assert 'filepath' in data.columns

    data = data.fillna(0)

    # Create a column group to group observations of the same release (identified by the commit)
    data['group'] = data.commit.astype('category').cat.rename_categories(range(1, data.commit.nunique() + 1))

    # Make sure the data is sorted by commit time (ascending)
    data.sort_values(by=['committed_at'], ascending=True)
    data = data.reset_index(drop=True)

    # Remove metadata
    data = data.drop(['commit', 'committed_at', 'filepath'], axis=1)

    X, y = data.drop(['failure_prone'], axis=1), data.failure_prone.values.ravel()

    return X, y


class DefectPredictor:

    def __init__(self):
        """
        Initialize a new DefectPredictor
        """

        self._balancers = []
        self._classifiers = []
        self._normalizers = []

        self.cv_report_map = dict()
        self.best_estimator = None
        self.best_estimator_average_precision = 0
        self.selected_features = list()

    @property
    def balancers(self):
        return self._balancers

    @balancers.setter
    def balancers(self, balancers: List[str]):
        for balancer in balancers:
            if balancer not in ('none', 'rus', 'ros'):
                raise ValueError(f'{balancer} is not supported')

            self._balancers.append(balancers_map[balancer])

    @property
    def normalizers(self):
        return self._normalizers

    @normalizers.setter
    def normalizers(self, normalizers: List[str]):
        for normalizer in normalizers:
            if normalizer not in ('none', 'minmax', 'std'):
                raise ValueError(f'{normalizer} is not supported')

            self._normalizers.append(normalizers_map[normalizer])

    @property
    def classifiers(self):
        return self._classifiers

    @classifiers.setter
    def classifiers(self, classifiers: List[str]):
        for classifier in classifiers:
            if classifier not in ('dt', 'logit', 'nb', 'rf', 'svm'):
                raise ValueError(f'{classifier} is not supported')

            self._classifiers = classifiers

    def train(self, data: pd.DataFrame) -> Pipeline:
        """
        Return the best fitted estimator, that is, the one that maximizes the average_precision
        :yield: the cross-validation report for every classifier
        :return: an estimator of type Pipeline
        """

        X, y = prepare_training_data(data)
        releases = X.group.tolist()
        X = X.drop(['group'], axis=1)

        scoring = dict(
            roc_auc='roc_auc',
            average_precision='average_precision',
            accuracy='accuracy',
            balanced_accuracy='balanced_accuracy',
            precision='precision',
            recall='recall',
            f1='f1',
            mcc=make_scorer(matthews_corrcoef)
        )

        for classifier in self.classifiers:
            estimator = classifiers_map[classifier]

            pipe = Pipeline([
                ('variance', VarianceThreshold(threshold=0)),  # Remove constant features
                ('balancing', None),  # To balance the training data See search_params['balancing'] below)
                ('normalization', None),  # To scale (and center) data. See search_params['normalization'] below
                # TODO feature_selection here
                ('classification', estimator)
            ])

            search_params = search_params_map[classifier]

            if self.balancers:
                search_params['balancing'] = self.balancers

            if self.normalizers:
                search_params['normalization'] = self.normalizers

            search = RandomizedSearchCV(pipe, search_params, cv=walk_forward_release(X, y, releases),
                                        scoring=scoring, refit='average_precision', verbose=0)

            search.fit(X, y)

            # Add additional metadata to the cv_results
            search.cv_results_['best_index_'] = search.best_index_

            buffer = io.StringIO()
            pd.DataFrame(search.cv_results_).to_json(buffer, orient='table', index=False)
            self.cv_report_map[classifier] = json.loads(buffer.getvalue())

            # Get the highest average_precision for this randomized search
            local_best_average_precision = search.cv_results_['mean_test_average_precision'][search.best_index_]

            if (not self.best_estimator) or (local_best_average_precision > self.best_estimator_average_precision):
                self.cv_report_map['best_classifier'] = classifier
                self.best_estimator = search.best_estimator_
                selected_features_indices = self.best_estimator.named_steps['variance'].fit(X).get_support(indices=True)
                self.selected_features = X.iloc[:, selected_features_indices].columns.tolist()

        return self.best_estimator

    def predict(self, unseen_data: pd.DataFrame) -> bool:
        """
        Predict an unseen instance as failure-prone or clean.
        :param unseen_data: pandas DataFrame containing the observation to predict
        :return: True if failure-prone. False, otherwise.
        """
        if not self.best_estimator:
            raise Exception('No model has been loaded yet. Please, load a model using instance.load(path_to_model_dir)')

        # Set missing features in unseen_data to zero
        for feature_name in self.selected_features:
            if feature_name not in unseen_data:
                unseen_data[feature_name] = 0

        # Select same model features
        unseen_data = unseen_data[np.intersect1d(unseen_data.columns, self.selected_features)]

        # Perform pre-process if any
        if self.best_estimator.named_steps['normalization']:
            unseen_data = pd.DataFrame(self.best_estimator.named_steps['normalization'].transform(unseen_data))

        clf = self.best_estimator.named_steps['classification']
        prediction = bool(clf.predict(unseen_data)[0])

        return prediction

    def load_model(self, path_to_dir: str):
        """
        :param path_to_dir: the path to the directory containing model-related files
        :return: None
        """
        model = joblib.load(os.path.join(path_to_dir, 'radondp_model.joblib'), mmap_mode='r')
        self.best_estimator = model['estimator']
        self.selected_features = model['selected_features']

    def dump_model(self, path_to_dir: str):
        """
        Dump the best model to file
        :param path_to_dir: the path to the directory containing model-related files
        :return: None
        """
        joblib.dump({'estimator': self.best_estimator,
                     'selected_features': self.selected_features,
                     'report': self.cv_report_map},
                    os.path.join(path_to_dir, 'radondp_model.joblib'))

    def dumps_model(self) -> bytes:
        """
        Dump the best model to a byte buffer
        :return: bytes
        """
        buff = io.BytesIO()
        joblib.dump({'estimator': self.best_estimator,
                     'selected_features': self.selected_features,
                     'report': self.cv_report_map},
                    buff)

        return buff.getvalue()



import numpy as np

decision_tree_params = dict(
    classification__criterion=['gini', 'entropy'],
    classification__splitter=['best', 'random'],
    classification__max_features=['auto', 'sqrt', 'log2', None],
    classification__class_weight=['balanced', None]
)

naive_params = dict(classification__var_smoothing=np.logspace(start=-10, stop=-8, num=10))

logistic_params = dict(
    classification__penalty=['l2', 'none'],
    classification__tol=np.logspace(-5, -3, 10),
    classification__C=np.linspace(0, 2, 10),
    classification__class_weight=[None, 'balanced'],
    classification__solver=['lbfgs', 'sag', 'saga'],
    classification__fit_intercept=[False, True]
)

max_depth = [int(x) for x in np.linspace(10, 110, num=11)]
max_depth.append(None)

random_forest_params = dict(
    classification__n_estimators=[int(x) for x in np.linspace(start=100, stop=2000, num=10)],
    classification__max_features=['auto', 'sqrt'],
    classification__max_depth=max_depth,
    classification__bootstrap=[True, False]
)

svc_params = dict(
    classification__C=[int(x) for x in np.linspace(start=1, stop=1000, num=100)],
    classification__gamma=['scale', 'auto'],
    classification__kernel=['rbf', 'sigmoid', 'poly'],
    classification__degree=[1, 2, 3, 4, 5],
    classification__shrinking=[False, True],
    classification__tol=[x for x in np.logspace(start=-5, stop=-1, num=10)],
    classification__class_weight=[None, 'balanced']
)

search_params_map = {
    'dt': decision_tree_params,
    'nb': naive_params,
    'logit': logistic_params,
    'rf': random_forest_params,
    'svm': svc_params
}



#!/usr/bin/env python

from setuptools import setup, find_packages

with open("requirements.txt", "r") as reqs_file:
    requirements = reqs_file.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = '0.2.5'

setup(name='radon-defect-predictor',
      version=VERSION,
      description='A Python library to train machine learning models for defect prediction of infrastructure code.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Stefano Dalla Palma',
      maintainer='Stefano Dalla Palma',
      author_email='stefano.dallapalma0@gmail.com',
      url='https://github.com/radon-h2020/radon-defect-predictor',
      download_url='https://github.com/radon-h2020/radon-defect-predictor/archive/{0}.tar.gz'.format(VERSION),
      packages=find_packages(exclude=('tests',)),
      entry_points={
          'console_scripts': ['radon-defect-predictor=radondp.cli:main'],
      },
      python_requires='>=3.6',
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "License :: OSI Approved :: Apache Software License",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Operating System :: OS Independent"
      ],
      insall_requires=requirements
)


import datetime
import json
import os
import pandas as pd
import requests

from ansiblemetrics import metrics_extractor as ansible_metrics_extractor
from toscametrics import metrics_extractor as tosca_metrics_extractor
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from zipfile import ZipFile

from .predictors import DefectPredictor


def valid_dir(x: str) -> str:
    """
    Check the directory exists
    :param x: a path
    :return: the path if exists; raise an ArgumentTypeError otherwise
    """
    if not os.path.isdir(x):
        raise ArgumentTypeError('Insert a valid path')

    return x


def valid_file(x: str) -> str:
    """
    Check the file exists
    :param x: a path
    :return: the path if exists; raise an ArgumentTypeError otherwise
    """
    if not os.path.isfile(x):
        raise ArgumentTypeError('Insert a valid path')

    return x


def valid_balancers(x: str):
    """
    Check x is a list of valid balancers
    :param x: a string representing a list of balancers (e.g., "none rus ros")
    :return: the list of balancers if every argument in x is a valid balancer; raise an ArgumentTypeError otherwise
    """
    balancers = x.split(' ')
    for balancer in balancers:
        if balancer not in ('none', 'rus', 'ros'):
            raise ArgumentTypeError(f'{balancer} is not a valid argument')

    return balancers


def valid_normalizers(x: str):
    """
    Check x is a list of valid normalizers
    :param x: a string representing a list of normalizers (e.g., "none minmax std")
    :return: the list of normalizers if every argument in x is a valid normalizer; raise an ArgumentTypeError otherwise
    """
    normalizers = x.split(' ')
    for normalizer in normalizers:
        if normalizer not in ('none', 'minmax', 'std'):
            raise ArgumentTypeError(f'{normalizer} is not a valid argument')

    return normalizers


def valid_classifiers(x: str):
    """
    Check x is a list of valid classifiers
    :param x: a string representing a list of classifiers (e.g., "dt logit nb rf svm")
    :return: the list of classifiers if every argument in x is a valid classifier; raise an ArgumentTypeError otherwise
    """
    classifiers = x.split(' ')
    for classifier in classifiers:
        if classifier not in ('dt', 'logit', 'nb', 'rf', 'svm'):
            raise ArgumentTypeError(f'{classifier} is not a valid argument')

    return classifiers


def set_train_parser(subparsers):
    parser = subparsers.add_parser('train', help='Train a brand new model from scratch')
    parser.add_argument(action='store',
                        dest='path_to_csv',
                        type=valid_file,
                        help='the path to the csv file containing the data for training')

    parser.add_argument(dest='classifiers',
                        type=valid_classifiers,
                        help='a list of classifiers to train. Possible choices [dt, logit, nb, rf, svm]')

    parser.add_argument('-b', '--balancers',
                        required=False,
                        dest='balancers',
                        type=valid_balancers,
                        help='a list of balancer to balance training data. Possible choices [none, rus, ros]')

    parser.add_argument('-n', '--normalizers',
                        required=False,
                        dest='normalizers',
                        type=valid_normalizers,
                        help='a list of normalizers to normalize data. Possible choices [none, minmax, std]')

    # TODO: add feature-selectors


    """
    parser.add_argument('--verbose',
                        action='store_true',
                        dest='verbose',
                        default=False,
                        help='show log')
    """


def set_predict_parser(subparsers):
    parser = subparsers.add_parser('predict', help='Predict unseen instances')

    parser.add_argument(action='store',
                        dest='language',
                        type=str,
                        choices=['ansible', 'tosca'],
                        help='the language of the file (i.e., TOSCA or YAML-based Ansible)')

    parser.add_argument(action='store',
                        dest='path_to_artefact',
                        type=valid_file,
                        help='the path to the artefact to analyze (i.e., an Ansible or Tosca file or .csar')


def set_download_model_parser(subparsers):
    parser = subparsers.add_parser('download-model', help='Download a pre-trained model from the online APIs')

    parser.add_argument(action='store',
                        dest='language',
                        type=str,
                        choices=['ansible', 'tosca'],
                        help='the language the model is trained on')

    parser.add_argument(action='store',
                        dest='host',
                        type=str,
                        choices=['github', 'gitlab'],
                        help='the platform the user\'s repository is hosted to')

    parser.add_argument(action='store',
                        dest='repository',
                        type=str,
                        help='the user\'s remote repository in the form <namespace>/<repository> (e.g., radon-h2020/radon-defect-prediction-cli)')


def get_parser():
    description = 'A Python library to train machine learning models for defect prediction of infrastructure code'

    parser = ArgumentParser(prog='radon-defect-predictor', description=description)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s 0.2.5')
    subparsers = parser.add_subparsers(dest='command')

    set_train_parser(subparsers)
    set_download_model_parser(subparsers)
    set_predict_parser(subparsers)

    return parser


def train(args: Namespace):
    dp = DefectPredictor()
    dp.balancers = args.balancers
    dp.normalizers = args.normalizers
    dp.classifiers = args.classifiers
    dp.train(pd.read_csv(args.path_to_csv))
    dp.dump_model(os.getcwd())
    exit(0)


def model(args: Namespace):
    """
        # Compute scores
        print('Downloading model...')
        scores = scorer.score_repository(
            path_to_repo=args.repository,
            full_name_or_id=args.repository_full_name_or_id,
            host=args.host
        )

        scores['commitFrequency'] = scores['commit_frequency']
        scores['coreContributors'] = scores['core_contributors']
        //scores['issueFrequency'] = scores['issue_frequency']
        scores['percentComments'] = scores['percent_comment']
        scores['percentIac'] = scores['iac_ratio']
        scores['sloc'] = scores['repository_size']

        // TODO: Add parameters to url query
    """

    url = f'https://radon-test-api.herokuapp.com/models/?language={args.language}&repository_size=100&return_model=1'
    r = requests.get(url, stream=True)
    if r.ok:
        dest = os.path.join(os.getcwd(), 'radondp_model.joblib')

        # For debug
        print("Saving to", dest)

        with open(dest, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024 * 8):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    os.fsync(f.fileno())
    else:  # HTTP status code 4XX/5XX
        print("Download failed: status code {}\n{}".format(r.status_code, r.text))

    exit(0)


def predict(args: Namespace):
    dp = DefectPredictor()
    dp.load_model(os.getcwd())

    report = []

    # Extract metrics
    if args.language == 'ansible':
        # Read content of the script to analyze
        with open(args.path_to_artefact, 'r') as f:
            script_content = f.read()
        unseen_data = pd.DataFrame(ansible_metrics_extractor.extract_all(script_content), index=[0])
        prediction = dp.predict(unseen_data)
        report.append(dict(
            file=args.path_to_artefact,
            failure_prone=prediction,
            analyzed_at=str(datetime.date.today())
        ))

    else:  # tosca

        if args.path_to_artefact.endswith('.csar'):

            with ZipFile(args.path_to_artefact, 'r') as zip_file:
                for filepath in zip_file.namelist():
                    if filepath.endswith('.tosca'):
                        try:
                            script_content = zip_file.read(filepath).decode('utf-8')
                            unseen_data = pd.DataFrame(tosca_metrics_extractor.extract_all(script_content), index=[0])
                            prediction = dp.predict(unseen_data)
                            report.append(dict(
                                file=os.path.join(args.path_to_artefact, filepath),
                                failure_prone=prediction,
                                analyzed_at=str(datetime.date.today())
                            ))
                        except ValueError:
                            pass
        else:
            # Read content of the script to analyze
            with open(args.path_to_artefact, 'r') as f:
                script_content = f.read()

            unseen_data = pd.DataFrame(tosca_metrics_extractor.extract_all(script_content), index=[0])

            prediction = dp.predict(unseen_data)
            report.append(dict(
                file=args.path_to_artefact,
                failure_prone=prediction,
                analyzed_at=str(datetime.date.today())
            ))

    if report:
        destination = os.path.join(os.getcwd(), 'radondp_predictions.json')
        if os.path.isfile(destination):
            with open(destination, 'r') as f:
                report.extend(json.load(f))

        with open(destination, 'w') as f:
            json.dump(report, f)

    exit(0)


def main():
    args = get_parser().parse_args()
    if args.command == 'train':
        train(args)
    elif args.command == 'download-model':
        model(args)
    elif args.command == 'predict':
        predict(args)
